﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using Windows.Devices.Enumeration;
using Windows.Devices.Spi;
using Windows.Devices.Gpio;

namespace PiPlates
{
    public sealed class DAQCplate:IDisposable
    {
        const byte MAXADDR = 8;
        const float DAQCversion = 1.02F;
        const byte GpioBaseAddr = 8;    //the DAQC has a base address of 8.  Other types of pi-plates have different base addresses (apparently)

        private float _vcc;
        private static GpioController _gpio;
        private static GpioPin _ppFrame;
        private static GpioPin _ppInt;
        private static SpiDevice _spi;
        private readonly byte _address;  //the address of this DAQCplate

        /// <summary>
        /// The address of this DAQC plate
        /// </summary>
        public byte Address => _address;

        static DAQCplate()
        {
            //initialize GPIO and SPI
            InitGPIO();
            InitSPI();
        }

        /// <summary>
        /// Creates a new instance of the DAQCPlate at the specified address
        /// </summary>
        /// <param name="addr">The address of the DAQCplate to create</param>
        public DAQCplate(byte addr)
        {
            if (addr > MAXADDR) throw new ArgumentOutOfRangeException(nameof(addr), $"address out of range - must be less than {MAXADDR - 1}");
            _address = addr;

            //initialize GPIO and SPI
            InitGPIO();
            InitSPI();
            
            //call getADDR to verify their is a DAQCplate installed at the specified address
            var rtn = getADDR(_address);   
            if (rtn - 8 == _address)
            {
                var ok = 0;
                //keep reading until the voltage is 'ok'
                while (ok == 0)
                {
                    //get the value of the Vcc (analog channel 8).  It should be at about 3.3V
                    _vcc = getADC(8);  
                        
                    //if Vcc value is greater than 3 volts then the board at this address is ready to use
                    if (_vcc > 3.0)   
                    {
                        ok = 1;
                    }

                    //todo: should we set some limit so we aren't in an infinite loop if the voltage is ever not greater than 3?
                }

                //set all outputs for this board to their zero state
                setDOUTall(0);
                setPWM(0, 0);
                setPWM(1, 0);
                setDAC(0, 0);
                setDAC(1, 0);
            }
            else
            {
                throw new Exception($"No DAQCplate found at address {_address}");
            }
        }

        /// <summary>
        /// Initialize the GPIO interface.
        /// </summary>
        private static void InitGPIO()
        {
            if (_gpio == null)
            {
                _gpio = GpioController.GetDefault();
                if (_gpio == null)
                {
                    throw new Exception("There is no GPIO controller on this device.");
                }
            }

            if (_ppFrame == null)
            {
                _ppFrame = _gpio.OpenPin(25, GpioSharingMode.Exclusive); //BCM 25
                _ppFrame.SetDriveMode(GpioPinDriveMode.Output);
            }

            if (_ppInt == null)
            {
                _ppInt = _gpio.OpenPin(22, GpioSharingMode.SharedReadOnly); //BCM 22
                //_ppInt.SetDriveMode(GpioPinDriveMode.Input);
                //_ppInt.DebounceTimeout = TimeSpan.FromMilliseconds(200);
            }
        }

        /// <summary>
        /// Initialize the SPI bus.
        /// </summary>
        private static void InitSPI()
        {
            if (_spi == null)
            {
                // Use chip select line CS1
                var settings = new SpiConnectionSettings(1)
                {
                    ClockFrequency = 500000,
                    Mode = SpiMode.Mode0,
                    SharingMode = SpiSharingMode.Shared
                };

                // Get a selector string for bus "SPI0"
                var aqs = SpiDevice.GetDeviceSelector("SPI0");

                // Find the SPI bus controller device with our selector string
                var dis = DeviceInformation.FindAllAsync(aqs).AsTask().ConfigureAwait(false).GetAwaiter().GetResult();
                if (dis.Count == 0) throw new Exception("There is no SPI bus 0 on this device.");

                // Create an SpiDevice with our bus controller and SPI settings
                _spi =  SpiDevice.FromIdAsync(dis[0].Id, settings).AsTask().ConfigureAwait(false).GetAwaiter().GetResult();
                if (_spi == null) throw new Exception("There is no SPI bus 0 on this device.");
            }
        }


        #region ADC Functions
        /// <summary>
        /// Returns the voltage from a single channel.
        /// </summary>
        /// <param name="channel">
        /// The analog channel number (0 - 8) you want to read voltage for.
        /// Channel 8 is the Vcc voltage.
        /// </param>
        /// <returns>The voltage of the specified channel (0-4.096)</returns>
        public float getADC(byte channel)
        {
            if (channel > 8) throw new ArgumentOutOfRangeException(nameof(channel), "channel out of range - must be less than 9");
            var resp = ppCMD(_address, 0x30, channel, 0, 2);
            float value = (256*resp[0] + resp[1]);
            value = (float)Math.Round(value*4.096/1024, 3);

            if (channel == 8) value = value * 2f;

            return value;
        }

        /// <summary>
        /// Returns voltage from all channels.
        /// </summary>
        /// <returns>An array of voltages where the index in the array corresponds to the channel.</returns>
        public float[] getADCall()
        {
            var resp = ppCMD(_address, 0x31, 0, 0, 16);
            var value = new float[8];
            for (var i = 0; i < value.Length; i++)
            {
                value[i] = 256*resp[2*i] + resp[2*i + 1];
                value[i] = (float) Math.Round(value[i]*4.096/1024, 3);
            }
            
            return value;
        }
        #endregion

        #region Digital Input Functions
        /// <summary>
        /// Returns a single bit value.
        /// </summary>
        /// <param name="bit">The bit number to read the value for (0 - 7)</param>
        /// <returns>The status of the digital input bit.  True = On, False = Off.</returns>
        public bool getDINbit(byte bit)
        {
            if (bit > 7) throw new ArgumentOutOfRangeException(nameof(bit), "bit argument out of range");
            var resp = ppCMD(_address, 0x20, bit, 0, 1);
            return resp[0] > 0;
        }
        
        /// <summary>
        /// Returns the value of all digital input bits
        /// </summary>
        /// <returns>A byte where each bit represents the current state of the corresponding digital input.</returns>
        public byte getDINall()
        {
            var resp = ppCMD(_address, 0x25, 0, 0, 1);


            return resp[0];
        }

        /// <summary>
        /// Enable interrupts for the input change on the specified bit.
        /// </summary>
        /// <param name="bit">The bit number you want to enable an interrupt for.</param>
        /// <param name="interuptEdge">The edge type that will trigger the interupt (Rising, Falling, or Both).</param>
        public void enableDINint(byte bit, InteruptEdgeType interuptEdge)
        {
            if (bit > 7) throw new ArgumentOutOfRangeException(nameof(bit), "bit argument out of range");
            switch (interuptEdge)
            {
                case InteruptEdgeType.Falling:
                    ppCMD(_address, 0x21, bit, 0, 0);
                    break;

                case InteruptEdgeType.Rising:
                    ppCMD(_address, 0x22, bit, 0, 0);
                    break;

                case InteruptEdgeType.Both:
                    ppCMD(_address, 0x23, bit, 0, 0);
                    break;
            }
        }

        /// <summary>
        /// Disable interrupts on the specified bit.
        /// </summary>
        /// <param name="bit">The bit number you want to disbale an interrupt for.</param>
        public void disableDINint(byte bit)
        {
            if (bit > 7) throw new ArgumentOutOfRangeException(nameof(bit), "bit argument out of range");

            ppCMD(_address, 0x24, bit, 0, 0);
        }

        /// <summary>
        /// a special function to read a DS18X20 temp sensor attached to the specified channel. 
        /// </summary>
        /// <remarks>Note that this function takes about 1 second to complete.</remarks>
        /// <param name="channel">The number of the specified bit input.</param>
        /// <param name="scale">Celsius, Fahrenheit, or Kelvin.</param>
        /// <returns>The temperature value converted to the desired scale.</returns>
        public float getTEMP(byte channel, TemperateUnits scale)
        {
            if (channel > 7) throw new ArgumentOutOfRangeException(nameof(channel), "channel value out of range");
            ppCMD(_address, 0x70, channel, 0, 0); // #initiate measurement
            Task.Delay(TimeSpan.FromSeconds(1)).Wait();
            var resp = ppCMD(_address, 0x71, channel, 0, 2); //   #get data

            var temp = resp[0]*256 + resp[1];
            if (temp > 0x8000)
            {
                temp = temp ^ 0xFFFF;
                temp = -(temp + 1);
            }
            var result = Math.Round(temp/16.0, 4);
            if (scale == TemperateUnits.Kelvin)
            {
                result = result + 273;
            }

            if (scale == TemperateUnits.Fahrenheit)
            {
                result = Math.Round(temp*1.8 + 32.2, 4);
            }

            return (float) result;
        }

        #endregion

        #region Hybrid Functions

        /// <summary>
        /// Reads distance from an HC-SR04 ultrasonic sensor.
        /// </summary>
        /// <param name="channel">The digital input / digital output pair required to interface to the sensor</param>
        /// <param name="units">Inches or Centimeters</param>
        /// <returns>The distance value converted to the desired units.</returns>
        public object getRANGE(byte channel, DistanceUnits units)
        {
            if (channel > 6) throw new ArgumentOutOfRangeException(nameof(channel), "channel value out of range");
            ppCMD(_address, 0x80, channel, 0, 0); //   #initiate measurement
            Task.Delay(TimeSpan.FromMilliseconds(1)).Wait();
            var resp = ppCMD(_address, 0x81, channel, 0, 2); //   #get data
            var range = resp[0]*256 + resp[1];
            if (range == 0) return "ERROR: sensor failure";
            float result = 0;

            if (units == DistanceUnits.Centimeters)
                result = range/58.326F;
            if (units == DistanceUnits.Inches)
                result = range/148.148F;
            result = (float) Math.Round(result, 2);
            return result;
        }

        #endregion

        #region LED Functions

        /// <summary>
        /// Turn ON one of the LEDs in the bicolor LED package
        /// </summary>
        /// <param name="color">The LED to turn ON.  0 = Red, 1 = Green.</param>
        public void setLED(byte color)
        {
            if (color > 1) throw new ArgumentOutOfRangeException(nameof(color), "Error: invalid LED value");
            ppCMD(_address, 0x60, color, 0, 0);
        }

        /// <summary>
        /// Turn OFF one of the LEDs in the bicolor LED package
        /// </summary>
        /// <param name="color">The LED to turn OFF.  0 = Red, 1 = Green.</param>
        public void clrLED(byte color)
        {
            if (color > 1) throw new ArgumentOutOfRangeException(nameof(color), "Error: invalid LED value");
            ppCMD(_address, 0x61, color, 0, 0);
        }

        /// <summary>
        /// Toggle the state of one of the LEDs in the bicolor LED package
        /// </summary>
        /// <param name="color">The LED to toggle.  0 = Red, 1 = Green.</param>
        public void toggleLED(byte color)
        {
            if (color > 1) throw new ArgumentOutOfRangeException(nameof(color), "Error: invalid LED value");
            ppCMD(_address, 0x62, color, 0, 0);
        }

        /// <summary>
        /// Get the current state of the specified LED
        /// </summary>
        /// <param name="color">The LED to toggle.  0 = Red, 1 = Green.</param>
        /// <returns>True if LED is on. false if LED is off.</returns>
        public bool getLED(byte color)
        {
            if (color > 1) throw new ArgumentOutOfRangeException(nameof(color), "Error: invalid LED value");
            var resp = ppCMD(_address, 0x63, color, 0, 1);
            return resp[0] > 0;
        }

        #endregion

        #region Switch Functions

        /// <summary>
        /// returns current state of on board switch. 
        /// </summary>
        /// <returns>True when switch is up, 0 when it is down.</returns>
        public bool getSWstate()
        {
            var resp = ppCMD(_address, 0x50, 0, 0, 1);
            return resp[0] > 0;
        }

        /// <summary>
        /// Allows the switch to generate an interrupt when pressed. Global interrupts must be enabled before using this function.
        /// </summary>
        public void enableSWint()
        {
            ppCMD(_address, 0x51, 0, 0, 0);
        }

        /// <summary>
        /// blocks switch on board from generating an interrupt.
        /// </summary>
        public void disableSWint()
        {
            ppCMD(_address, 0x52, 0, 0, 0);
        }

        /// <summary>
        /// Pushing button on ppGPIO will short RPI GPIO23 to GND and then remove 5VDC 45 seconds later. 
        /// </summary>
        /// <remarks>Note that this setting is saved in nonvolatile memory and only has to be performed once</remarks>
        public void enableSWpower()
        {
            ppCMD(_address, 0x53, 0, 0, 0);
        }

        /// <summary>
        /// Disables the on-board switch as a power button.
        /// </summary>
        /// <remarks>Note that this setting is saved in nonvolatile memory and only has to be performed once</remarks>
        public void disableSWpower()
        {
            ppCMD(_address, 0x54, 0, 0, 0);
        }

        #endregion

        #region Digital Output Functions

        /// <summary>
        /// Turns ON a single bit
        /// </summary>
        /// <param name="bit">The bit that you want to turn ON. (0 - 6)</param>
        public void setDOUTbit(byte bit)
        {
            if (bit > 6) throw new ArgumentOutOfRangeException(nameof(bit), "ERROR: bit argument out of range - must be less than 7");
            ppCMD(_address, 0x10, bit, 0, 0);
        }

        /// <summary>
        /// Turns OFF a single bit
        /// </summary>
        /// <param name="bit">The bit that you want to turn OFF. (0 - 6)</param>
        public void clrDOUTbit(byte bit)
        {
            if (bit > 6) throw new ArgumentOutOfRangeException(nameof(bit), "ERROR: bit argument out of range - must be less than 7");
            ppCMD(_address, 0x11, bit, 0, 0);
        }

        /// <summary>
        /// Toggles a single bit
        /// </summary>
        /// <param name="bit">The bit that you want to toggle. (0 - 6)</param>
        public void toggleDOUTbit(byte bit)
        {
            if (bit > 6) throw new ArgumentOutOfRangeException(nameof(bit), "ERROR: bit argument out of range - must be less than 7");
            ppCMD(_address, 0x12, bit, 0, 0);
        }

        /// <summary>
        /// Sets all digital outputs based on the corresponding bits in the specified value.
        /// </summary>
        /// <param name="value">The value to set the digital outputs to.  Each bit represents a digital channel.</param>
        /// <remarks>Because only bits 0 - 6 are available, the value must be between 0 and 127.</remarks>
        public void setDOUTall(byte value)
        {
            if (value > 127) throw new ArgumentOutOfRangeException(nameof(value), "ERROR: value argument out of range - must be less than 128");
            ppCMD(_address, 0x13, value, 0, 0);
        }

        /// <summary>
        /// Gets the value of all of the digital outputs
        /// </summary>
        /// <returns>Returns a byte where each bit represents the current state of the corresponding digital output.</returns>
        public byte getDOUTall()
        {
            var resp = ppCMD(_address, 0x14, 0, 0, 1);

            return resp[0];
        }

        #endregion

        #region PWM and DAC Output Functions

        /// <summary>
        /// set PWM duty cycle from 0 to 1023 (0 to 100%)
        /// </summary>
        /// <param name="channel">The PWN channel to set.</param>
        /// <param name="value">The duty cycle. (0 = 0%, 1023 = 100%)</param>
        public void setPWM(byte channel, byte value)
        {
            if (channel > 1) throw new ArgumentOutOfRangeException(nameof(channel), "Error: PWM channel must be 0 or 1");
            byte hibyte = (byte) (value >> 8);
            byte lobyte = (byte) (value - (hibyte << 8));
            ppCMD(_address, (byte) (0x40 + channel), hibyte, lobyte, 0);
        }

        /// <summary>
        /// Gets the current duty cycle of the specified PWM output.
        /// </summary>
        /// <param name="channel">The PWM output channel (0 or 1)</param>
        /// <returns>The duty cycle. (0 = 0%, 1023 = 100%)</returns>
        public float getPWM(byte channel)
        {
            if (channel > 1) throw new ArgumentOutOfRangeException(nameof(channel), "Error: PWM channel must be 0 or 1");
            var resp = ppCMD(_address, (byte) (0x40 + channel + 2), 0, 0, 2);
            var value = 256*resp[0] + resp[1];
            return value;
        }

        /// <summary>
        /// Set the output voltage of the specified DAC channel.
        /// </summary>
        /// <param name="channel">The DAC output channel.</param>
        /// <param name="value">The output voltage to set the DAC to (0V - 4.097V)</param>
        public void setDAC(byte channel, float value)
        {
            if (channel > 1) throw new ArgumentOutOfRangeException(nameof(channel), "Error: ADC channel must be 0 or 1");
            if (value > 4.097)
                throw new ArgumentOutOfRangeException(nameof(value), "ERROR: ADC argument out of range - must be less than 4.097 volts");
            var tmp = (int) (value/_vcc*1024);
            var hibyte = (byte) (tmp >> 8);
            var lobyte = (byte) (tmp - (hibyte << 8));
            ppCMD(_address, (byte) (0x40 + channel), hibyte, lobyte, 0);
        }

        /// <summary>
        /// Returns the current DAC output voltage
        /// </summary>
        /// <param name="channel">The DAC channel to return.</param>
        /// <returns>The current voltage of the specified DAC channel (0V - 4.097V)</returns>
        public float getDAC(byte channel)
        {
            if (channel > 1) throw new ArgumentOutOfRangeException(nameof(channel), "Error: ADC channel must be 0 or 1");
            var resp = ppCMD(_address, (byte) (0x40 + channel + 2), 0, 0, 2);
            var value = 256*resp[0] + resp[1];
            var result = value*_vcc/1023;
            return result;
        }

        /// <summary>
        /// Calibrate the DAC channels
        /// </summary>
        public void calDAC()
        {
            var rtn = getADDR(_address);
            if (rtn - 8 == _address)
            {
                _vcc = getADC(8);
            }
            else
            {
                _vcc = 10000; //todo: why 10,000?                
            }
        }

        #endregion

        #region Interrupt Control Functions

        /// <summary>
        /// enable interrupts from the DAQC. GPIO22 will be pulled low if an enabled event occurs.
        /// </summary>
        public void intEnable()
        {
             ppCMD(_address, 0x04, 0, 0, 0);
            _ppInt.ValueChanged += InteruptOccurred; //When the GPIO pin changes at least one interupt occurred.
        }
        
        /// <summary>
        /// disables and clears all interrupts on the DAQC.
        /// </summary>
        public void intDisable()
        {
            ppCMD(_address, 0x05, 0, 0, 0);
            _ppInt.ValueChanged -= InteruptOccurred; // ignore any changes to the GPIO pin
        }

        /// <summary>
        /// returns 16 bit flag value then clears all INT flags
        /// </summary>
        /// <returns>16bit flag value.</returns>
        public short getINTflags()
        {
            var resp = ppCMD(_address, 0x06, 0, 0, 2);
            var value = 256*resp[0] + resp[1];
            return (short) value;
        }


        private void InteruptOccurred(GpioPin sender, GpioPinValueChangedEventArgs args)
        {
            //GPIO22 will be pulled low if an enabled event occurs.  Ignore rising edge
            if (args.Edge != GpioPinEdge.FallingEdge) return;

            //get the interupt flags to see which input generated the interupt
            var flags = getINTflags();

            Task.Delay(TimeSpan.FromSeconds(.001)).Wait();  //

            //check bits 0 - 7 to see if one of the digital inputs caused the interupt
            for (byte i = 0; i < 8; i++)
            {
                var isInterupt = (flags & 1 << i) > 0;
                if (isInterupt)
                {
                    //raise event that a DigitalInput Interupt Occurred
                    OnDigitalInputInteruptOccurred(i, getDINbit(i));
                }
            }

            //bit 8 is the switch interupt.  It only occurs when the switch is pressed, not when the switch is released.
            if ((flags & 1 << 8) > 0)
            {
                OnSwitchInteruptOccurred();
            }
        }

        public event EventHandler<DigitalInputInteruptOccurredEventArgs> DigitalInputInteruptOccurred;
        public event EventHandler SwitchInteruptOccurred;
        #endregion

        #region System Functions

        /// <summary>
        /// returns the hardware revision of the pi-plate
        /// </summary>
        /// <returns>The hardware revision.</returns>
        public float getHWrev()
        {
            var resp = ppCMD(_address, 0x02, 0, 0, 1);
            var rev = resp[0];
            var whole = (float) (rev >> 4);
            var point = (float) (rev & 0x0F);
            return (float) (whole + point/10.0);
        }

        /// <summary>
        /// returns the firmware revision of the pi-plate
        /// </summary>
        /// <returns>The firmware revision.</returns>
        public float getFWrev()
        {
            var resp = ppCMD(_address, 0x03, 0, 0, 1);
            var rev = resp[0];
            var whole = (float) (rev >> 4);
            var point = (float) (rev & 0x0F);
            return (float) (whole + point/10.0);
        }

        /// <summary>
        /// Gets the version number of this library
        /// </summary>
        /// <returns>The version of this module</returns>
        public float getVersion()
        {
            return DAQCversion;
        }

        /// <summary>
        /// Return address of pi-plate. Used for polling available boards at power up.
        /// </summary>
        /// <remarks>Note DAQCplates are between 8-15</remarks>
        /// <param name="addr">The DAQCplate address (0-7)</param>
        /// <returns>Address of the pi-plate.</returns>
        private static int getADDR(byte addr)
        {
            var resp = ppCMD(addr, 0x00, 0, 0, 1);
            return resp[0];
        }

        /// <summary>
        /// return Pi-Plate descriptor string
        /// </summary>
        /// <returns></returns>
        public string getID()
        {
            var addr = _address + GpioBaseAddr;

            if (addr > 255)
                throw new ArgumentOutOfRangeException($"address out of range - must be less than 255");
            var id = "";
            var arg = new byte[4];
            arg[0] = (byte) addr;
            arg[1] = 0x1;
            arg[2] = 0;
            arg[3] = 0;

            _ppFrame.Write(GpioPinValue.High);
            _spi.Write(arg);
            var count = 0;
            Task.Delay(TimeSpan.FromSeconds(.0001)).Wait();

            while (count < 20)
            {
                var dummy = new byte[1];
                _spi.ConnectionSettings.ClockFrequency = 500000;
                _spi.TransferFullDuplex(new byte[] {00}, dummy);

                Task.Delay(TimeSpan.FromSeconds(.0001)).Wait();
                if (dummy[0] != 0)
                {
                    var num = dummy[0];
                    id = id + char.ConvertFromUtf32(num);
                    count = count + 1;
                }
                else
                {
                    count = 20;
                }
            }
            _ppFrame.Write(GpioPinValue.Low);
            return id;
        }

        /// <summary>
        /// Read a byte of data from program memory
        /// </summary>
        /// <param name="paddr">The block address of the program memory data to read.</param>
        /// <returns></returns>
        public byte getPROGdata(byte paddr)
        {
            var resp = ppCMD(_address, 0xF0, (byte) (paddr >> 8), (byte) (paddr & 0xFF), 2);
            var value = 256*resp[0] + resp[1];
            return (byte) value;
        }

        /// <summary>
        /// Determine what Pi-Plates are on stack as well as their addresses
        /// </summary>
        /// <returns></returns>
        public static int[] Poll()
        {
            var ppFoundCount = 0;
            var result = new List<int>();
            for (byte i = 0; i < 8; i++)
            {
                var rtn = getADDR(i);
                if ((rtn - 8) == i)
                {
                    result.Add(rtn - 8);
                    ppFoundCount += 1;
                }
            }

            return ppFoundCount == 0 ? null : result.ToArray();
        }

        /// <summary>
        /// Sends the SPI command to the device at the specified address
        /// </summary>
        /// <param name="addr">The address of the DAQCplate</param>
        /// <param name="cmd">The command to send</param>
        /// <param name="param1">command parameter 1</param>
        /// <param name="param2">command parameter 2</param>
        /// <param name="bytes2Return">The number of bytes expected to be returned</param>
        /// <returns>A byte[] with the results of the command</returns>
        private static byte[] ppCMD(byte addr, byte cmd, byte param1, byte param2, int bytes2Return)
        {
            var arg = new byte[4];
            var resp = new byte[bytes2Return];
            arg[0] = (byte) (addr + GpioBaseAddr);
            arg[1] = cmd;
            arg[2] = param1;
            arg[3] = param2;
            //Task.Delay(TimeSpan.FromSeconds(.0005)).Wait();     //# time.sleep(.0005)
            _ppFrame.Write(GpioPinValue.High);
            _spi.Write(arg);
            if (bytes2Return > 0)
            {
                Task.Delay(TimeSpan.FromSeconds(.0001)).Wait();
                _spi.ConnectionSettings.ClockFrequency = 500000;
                for (int i = 0; i < bytes2Return; i++)
                {
                    var dummy = new byte[1];
                    _spi.TransferFullDuplex(new byte[] {00}, dummy);
                    resp[i] = dummy[0];
                }
            }
            _ppFrame.Write(GpioPinValue.Low);

            return resp;
        }

        #endregion

        private void OnDigitalInputInteruptOccurred(byte channel, bool value)
        {
            DigitalInputInteruptOccurred?.Invoke(this, new DigitalInputInteruptOccurredEventArgs(channel, value));
        }

        private void OnSwitchInteruptOccurred()
        {
            SwitchInteruptOccurred?.Invoke(this, EventArgs.Empty);
        }

        public void Dispose()
        {
            _spi.Dispose();
            _ppInt.Dispose();
            _ppFrame.Dispose();
        }
    }
}
